#
# Virtuozzo containers hauler module
#

import logging
import os
import shlex
import subprocess
import util
import signal
import select
import glob

import pycriu.rpc_pb2

import criu_api
import criu_cr
import fs_haul_ploop
import pycriu


vz_global_conf = "/etc/vz/vz.conf"
vz_conf_dir = "/etc/vz/conf/"
vzctl_bin = "vzctl"
ip_bin = "ip"
cgget_bin = "cgget"
cgexec_bin = "cgexec"


vz_cgroup_mount_map = {
	"cpu,cpuacct": "cpu",
	"cpuset": "cpuset",
	"net_cls": "net_cls",
	"memory": "memory",
	"devices": "devices",
	"blkio": "blkio",
	"freezer": "freezer",
	"perf_event": "perf_event",
	"hugetlb": "hugetlb",
	"name=systemd": "systemd",
	"pids": "pids",
}

cgroup_controllers = [
	"cpuacct,cpu",
	"cpuset",
	"net_cls",
	"net_prio,net_cls",
	"memory",
	"devices",
	"blkio",
	"freezer",
	"perf_event",
	"hugetlb",
	"name=systemd",
	"pids",
	"beancounter",
	"ve",
]

cg_root_controllers = [
	"name=systemd",
	"ve",
	"beancounter",
]


vz_action_scripts = {
	"post-network-lock": ["/usr/libexec/criu/scripts/nfs-ports-allow.sh"],
}


class p_haul_type(object):
	def __init__(self, ctid):
		self._ctid = ctid
		self._ct_priv = ""
		self._ct_root = ""
		#
		# This list would contain (v_in, v_out, v_br) tuples where
		# v_in is the name of veth device in CT
		# v_out is its peer on the host
		# v_bridge is the bridge to which thie veth is attached
		#
		self._veths = []
		self.__verbose = criu_api.def_verb
		self._args_path = ''
		self._resume_wait_pipe = []
		self._vzctl_proc = 0
		self._has_nfs = None
		self._has_net_disk = None
		self._bindmounts = []
		self._disks = []
		self.nfsd_alive_services = []

	def __load_ct_config(self, path):
		logging.info("Loading config file from %s", path)

		# Read container config
		with open(self.__ct_config_path(path)) as ifd:
			config = _parse_vz_config(ifd.read())

		# Read global config
		with open(vz_global_conf) as ifd:
			global_config = _parse_vz_config(ifd.read())

		# Extract veth pairs, later we will equip restore request with this
		# data and will use it while (un)locking the network
		if "NETIF" in config:
			v_in, v_out, v_bridge = None, None, None
			for parm in config["NETIF"].split(","):
				pa = parm.split("=")
				if pa[0] == "ifname":
					v_in = pa[1]
				elif pa[0] == "host_ifname":
					v_out = pa[1]
				elif pa[0] == "bridge":
					v_bridge = pa[1]
			if v_in and v_out:
				logging.info("\tCollect %s -> %s (%s) veth",
							v_in, v_out, v_bridge)
				self._veths.append(util.net_dev(v_in, v_out, v_bridge))

		# Extract private path from config
		if "VE_PRIVATE" in config:
			self._ct_priv = _expand_veid_var(config["VE_PRIVATE"], self._ctid)
		else:
			self._ct_priv = _expand_veid_var(
				global_config["VE_PRIVATE"], self._ctid)

		# Extract root path from config
		if "VE_ROOT" in config:
			self._ct_root = _expand_veid_var(config["VE_ROOT"], self._ctid)
		else:
			self._ct_root = _expand_veid_var(
				global_config["VE_ROOT"], self._ctid)

		if "BINDMOUNT" in config:
			# bindmounts are space-separated, spaces are not supported in paths
			# must be safe to use split() here
			for bindmount in config["BINDMOUNT"].split():
				# cutting off bindmount options here
				bm_paths = bindmount.split(",")[0]

				# we're only interested in "src:dst" bindmounts, "dst"-only are skipped
				if ":" in bm_paths:
					src, dst = bm_paths.split(":")
					self._bindmounts.append((src, dst))

	def __load_ct_config_dst(self, path):
		if not os.path.isfile(self.__ct_config_path(path)):
			raise Exception("CT config missing on destination")
		self.__load_ct_config(path)

	def __ct_config_path(self, conf_dir):
		return os.path.join(conf_dir, "{0}.conf".format(self._ctid))

	def __cg_set_veid(self):
		"""Initialize veid in ve.veid for ve cgroup"""

		veid_path = "/sys/fs/cgroup/ve/{0}/ve.veid".format(self._ctid)
		with open(veid_path, "w") as f:
			if self._ctid.isdigit():
				veid = self._ctid
			else:
				veid = str(int(self._ctid.partition("-")[0], 16))
			f.write(veid)

	def __has_nfs_mount(self):
		"""Check if NFS mount is present in the container"""
		minfo_f = "/proc/{0}/mountinfo".format(self.root_task_pid())
		with open(minfo_f, "r") as f:
			minfo = f.readlines()

		# Mountinfo entry example:
		# 36 35 98:0 /mnt1 /mnt2 rw,noatime master:1 - ext3 /dev/root rw,errors=continue
		# (1)(2)(3)   (4)   (5)      (6)      (7)   (8) (9)   (10)         (11)
		# (8) is a separator, (7) is a space-separated list of optional fields
		# More info in man proc(5)

		for entry in minfo:
			p1, p2 = entry.split(" - ")
			fstype = p2.split()[0]
			if "nfs" in fstype:
				return True
		return False

	def __has_nfsd_feature(self):
		"""Check if NFSD feature is enabled for the container"""
		VE_FEATURE_NFSD = 1 << 8
		proc = subprocess.Popen(
			[cgget_bin, "-n", "-v", "-r", "ve.features", self._ctid],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		features = proc.communicate()[0]
		if proc.returncode == 0:
			return features and VE_FEATURE_NFSD
		else:
			logging.warning("cgget failed to get ve features")
			return False

	def has_k8s_enabled(self):
		"""Check if K8S support is enabled on a server"""
		logging.info("Checking k8s support")
		p = subprocess.Popen(["systemctl", "status", "vz-k8s-inside-ct"])
		p.communicate()
		result = (p.returncode == 0)
		logging.info("\t- k8s support: {0}".format(result))
		return result

	def init_src(self):
		self._fs_mounted = True
		self._bridged = True
		self.__load_ct_config(vz_conf_dir)

	def init_dst(self):
		self._fs_mounted = False
		self._bridged = False
		self.__load_ct_config_dst(vz_conf_dir)

	def set_options(self, opts):
		self.__verbose = opts["verbose"]
		self.__external_disks = opts.get("vz_external_disks")
		self._disks = self.__parse_external_disks_arg(self.__external_disks)

	def adjust_criu_req(self, req):
		"""Add module-specific options to criu request"""
		# Specify dump specific options
		if req.type == pycriu.rpc.DUMP:

			# Specify root fs
			req.opts.root = self._ct_root

			# Restore cgroups configuration
			req.opts.manage_cgroups = True

			# Setup mapping for external Virtuozzo specific cgroup mounts
			self.__fill_vz_cgroup_ext_mount_map(req)

			# Setup cgroup root
			self.__fill_cgroup_root(req)

			# Setup bindmounts map
			self.__fill_bindmounts_map(req)

			# Specify "external" ploop disks and partitions to CRIU
			# Starting with criu-3.15.1.19 we report even the root ploop disk, thus "external disks" option contains all disks
			for uuid, major, minor in self._disks:
				req.opts.external.append(
					"dev[{0}/{1}]:{2}".format(major, minor, uuid))

			# Increase ghost-limit up to 4Gb
			req.opts.ghost_limit = 4 << 30 - 1

			# Configure cgroup props file
			req.opts.cgroup_props_file = "/usr/libexec/libvzctl/scripts/cgroup-props.conf"

		# Specify both predump and dump specific options
		if req.type == pycriu.rpc.PRE_DUMP or req.type == pycriu.rpc.DUMP:

			# Specify ve for criu
			req.opts.vz_ve = self._ctid

			# Specify freezer cgroup
			req.opts.freeze_cgroup = \
				"/sys/fs/cgroup/freezer/machine.slice/{0}/".format(self._ctid)

			# Setup cgroup dump controllers
			for cgname in cgroup_controllers:
				req.opts.cgroup_dump_controller.append(cgname)

			# Increase timeout up to 180 seconds
			req.opts.timeout = 180

	def __fill_vz_cgroup_ext_mount_map(self, req):
		with open("/proc/self/mountinfo") as x:
			mountinfo = x.readlines()
		# Walk cgroups maps
		for cgkey in vz_cgroup_mount_map:
			# Iterate through lines in mountinfo until we find a match
			for buf in mountinfo:
				cg_matches = True

				# Records in mountinfo might have different amount of columns due to optional fields.
				# To work properly with varying length we must split line by " - " separator and
				# process it according to kernel/Documentation/filesystems/proc.txt
				pre_s, post_s = buf.split(" - ")
				pre_s_arr = pre_s.split()
				post_s_arr = post_s.split()

				if post_s_arr[0] != "cgroup":
					continue

				# If key consists of two cgroups (e.g."cpu,cpuacct"), search each
				for cgname in cgkey.split(","):
					if cgname not in post_s_arr[2].split(","):
						cg_matches = False
						break

				if cg_matches:
					req.opts.ext_mnt.add(key = pre_s_arr[4],
							val = vz_cgroup_mount_map[cgkey])
					break

	def __fill_cgroup_root(self, req):
		global_cgroot = "/machine.slice/{0}".format(self._ctid)
		cg_path = "/{0}".format(self._ctid)

		# add global cg root, controller is not specified for global cg_root
		req.opts.cg_root.add(path = global_cgroot)

		# fill in cgroups
		for cgroot in cg_root_controllers:
			req.opts.cg_root.add(ctrl = cgroot,
					path = cg_path)

	def __fill_bindmounts_map(self, req):
		for src, dst in self._bindmounts:
			req.opts.ext_mnt.add(key = dst, val = src)

	def root_task_pid(self):
		path = "/var/run/ve/{0}.init.pid".format(self._ctid)
		with open(path) as pidfile:
			pid = pidfile.read()
			return int(pid)

	def get_meta_images(self, path):
		return []

	def put_meta_images(self, path):
		pass

	def ve_sync(self, timeout, fail_on_timeout):
		"""Perform sync in a container context"""
		logging.info("\tSync VE filesystems")
		signal.signal(signal.SIGALRM, util.alarm_handler)
		signal.alarm(timeout)
		try:
			proc = subprocess.Popen(
				[cgexec_bin, "-g", "ve,beancounter:{0}".format(self._ctid), "/bin/sync"],
				stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
			proc_output = proc.communicate()[0]
			signal.alarm(0)
		except util.Alarm:
			if fail_on_timeout:
				raise Exception("VE filesystems sync timed out ({0} seconds)".format(timeout))
			else:
				logging.warning("VE filesystems sync timed out (non-blocking): ({0} seconds)".format(timeout))
				proc.kill()
		if proc.returncode:
			raise Exception("VE filesystems sync failed ({0}):{1}".format(proc.returncode, proc_output))

	def final_dump(self, pid, img, ccon, fs):
		"""Perform Virtuozzo-specific final dump"""
		criu_cr.criu_dump(self, pid, img, ccon, fs)
		self.__post_final_dump(img)

	def __pre_final_dump(self, img):
		"""Sync ve and create extra images before final dump"""
		assert self._has_nfs != None and self._has_net_disk != None

		if self._has_nfs or self._has_net_disk:
			self.ve_sync(300, False)

		extra_images = (
			("vz_clock_bootbased.img", "ve.clock_bootbased"),
			("vz_clock_monotonic.img", "ve.clock_monotonic"),
			("vz_iptables_mask.img", "ve.iptables_mask"),
			("vz_os_release.img", "ve.os_release"),
			("vz_features.img", "ve.features"),
			("vz_aio_max_nr.img", "ve.aio_max_nr"),
			("vz_pid_max.img", "ve.pid_max"))
		for image_name, var_name in extra_images:
			self.__create_cgget_extra_image(img, image_name, var_name)

		if self.__has_nfsd_feature():
			logging.info("\t\tCreate extra images for the nfs-server")
			self.__dump_nfs_server(img, "vz_nfsd_active.img")

	def __revert_pre_dump(self, img):
		""" Revert any inconsistencies we had introduced before dump """
		self.__restore_nfs_server(img, "vz_nfsd_active.img")

	def __service_stop_and_save_state(self, service, img, image_name):
		""" Stop service if alive and save state """
		systemctl_bin = "/bin/systemctl"
		timeout_bin = "/usr/bin/timeout"

		proc = subprocess.Popen(
			[timeout_bin, "10s", vzctl_bin, "--skipowner", "--skiplock", "exec3", self._ctid, systemctl_bin, "is-active", service],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc.communicate()

		if proc.returncode == 0:
			self.nfsd_alive_services.append(service)
			self.__create_extra_image(img, image_name, "\n".join(self.nfsd_alive_services))
			proc = subprocess.Popen(
				[timeout_bin, "10s", vzctl_bin, "--skipowner", "--skiplock", "exec3", self._ctid, systemctl_bin, "stop", service],
				stdout=subprocess.PIPE,	stderr=subprocess.STDOUT)
			proc_output = proc.communicate()[0]
			if proc.returncode:
				raise Exception("Failed to stop %s: %s", service, proc_output)
			logging.info("Stopped %s service", service)

	def __dump_nfs_server(self, img, image_name):
		"""Dump nfs-server active status and stop it if necessary"""
		self.nfsd_alive_services = []
		self.__create_extra_image(img, image_name, "")
		for service in ["nfs-server", "nfs-mountd", "nfsdcld", "nfs-blkmap"]:
			self.__service_stop_and_save_state(service, img, image_name)

	def __restore_nfs_server(self, img, image_name):
		"""Revert nfs-server active status """
		systemctl_bin = "/bin/systemctl"
		timeout_bin = "/usr/bin/timeout"

		if not len(self.nfsd_alive_services):
			return

		for service in self.nfsd_alive_services:
			proc = subprocess.Popen(
				[timeout_bin, "10s", vzctl_bin, "--skipowner", "--skiplock", "exec3", self._ctid, systemctl_bin, "start", service],
				stdout=subprocess.PIPE,	stderr=subprocess.STDOUT)
			proc_output = proc.communicate()[0]
			if proc.returncode:
				raise Exception("Failed to start %s: %s", service, proc_output)
			logging.info("Restarted %s service", service)
		logging.info("Restarted nfs server related services")

	def __post_final_dump(self, img):
		"""Sync ve and create extra images after final dump"""
		if self._has_nfs or self._has_net_disk:
			self.ve_sync(10, True)

		extra_images = (
			("vz_core_pattern.img", ["cat", "/proc/sys/kernel/core_pattern"]),)
		for image_name, exec_args in extra_images:
			self.__create_cgexec_extra_image(img, image_name, exec_args)

	def __create_cgget_extra_image(self, img, image_name, var_name):
		"""Create extra image using cgget output"""
		proc = subprocess.Popen(
			[cgget_bin, "-n", "-v", "-r", var_name, self._ctid],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		image_data = proc.communicate()[0]
		if proc.returncode == 0:
			self.__create_extra_image(img, image_name, image_data)
		else:
			logging.warning("cgget failed to create %s", image_name)

	def __create_cgexec_extra_image(self, img, image_name, exec_args):
		"""Create extra image using cgexec output"""
		proc = subprocess.Popen(
			[cgexec_bin, "-g", "ve:{0}".format(self._ctid)] + exec_args,
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		image_data = proc.communicate()[0]
		if proc.returncode == 0:
			self.__create_extra_image(img, image_name, image_data)
		else:
			logging.warning("cgexec failed to create %s", image_name)

	def __create_extra_image(self, img, image_name, image_data):
		image_path = os.path.join(img.image_dir(), image_name)
		with open(image_path, "w") as f:
			f.write(image_data)

	def _finalize_restore(self):
		"""Finalize restore"""
		if self._has_nfs:
			logging.info("\tContinue resume process")
			os.write(self._resume_wait_pipe[1], "\0\0\0\0")

		proc_output = self._vzctl_proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		try:
			if self._vzctl_proc.returncode != 0:
				raise Exception("Restore failed ({0}):{1}".format(self._vzctl_proc.returncode, self._vzctl_proc.communicate()[0]))
		finally:
			# Remove restore extra arguments
			self.__remove_restore_extra_args(self._args_path)

	def final_restore(self, img, connection):
		"""Perform Virtuozzo-specific final restore"""
		# Setup restore extra arguments
		self._args_path = os.path.join(img.image_dir(), "restore-extra-args")
		self.__setup_restore_extra_args(self._args_path, img, connection)

		# Prepare wait pipes for action scripts
		ph_env = os.environ.copy()
		if self._has_nfs:
			ph_pipe_wait, vz_pipe_write = os.pipe()
			self._resume_wait_pipe = os.pipe()
			ph_env["CRIU_ACTION_POST_RESUME_READ_FD"] = str(vz_pipe_write)
			ph_env["CRIU_ACTION_POST_RESUME_WRITE_FD"] = str(self._resume_wait_pipe[0])

		# Run vzctl restore
		logging.info("Starting vzctl restore")
		self._vzctl_proc = subprocess.Popen([vzctl_bin, "--skipowner", "--skiplock", "--ignore-ha-cluster",
			"restore", self._ctid, "--skip_arpdetect",
			"--dumpfile", img.image_dir()], env=ph_env,
			stdout=subprocess.PIPE,
			stderr=subprocess.STDOUT)

		if self._has_nfs:
			# Close write end of the pipe to make sure pipe dies if child exits
			# so we can poll just the pipe, instead of checking both child and pipe
			os.close(vz_pipe_write)
			os.close(self._resume_wait_pipe[0])
			select.select([ph_pipe_wait], [], [])

			if self._vzctl_proc.poll():
				self.__remove_restore_extra_args(self._args_path)
				raise Exception("Restore failed ({0}):{1}".format(self._vzctl_proc.returncode, self._vzctl_proc.communicate()[0]))
			os.read(ph_pipe_wait, 4)
			logging.info("\tReached post-restore stage, kill source container early for NFS scenario")
		else:
			self._finalize_restore()

	def __setup_restore_extra_args(self, path, img, connection):
		"""Create temporary file with extra arguments for criu restore"""
		extra_args = [
			"VE_WORK_DIR={0}\n".format(img.work_dir()),
			"VE_RESTORE_LOG_PATH={0}\n".format(
				connection.get_log_name(pycriu.rpc.RESTORE)),
			"VE_CRIU_LOGLEVEL={0}\n".format(self.__verbose)]
		with open(path, "w") as f:
			f.writelines(extra_args)

	def __remove_restore_extra_args(self, path):
		"""Remove temporary file with extra arguments for criu restore"""
		if os.path.isfile(path):
			os.remove(path)

	def prepare_ct(self, pid):
		"""Create cgroup hierarchy and put root task into it."""
		self.__cg_set_veid()

	def mount(self):
		logging.info("Mounting CT root to %s", self._ct_root)
		logging.info("Running vzctl mount")
		proc = subprocess.Popen(
			[vzctl_bin, "--skipowner", "--skiplock", "mount", self._ctid],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		self._fs_mounted = True
		return self._ct_root

	def umount(self):
		if self._fs_mounted:
			logging.info("Umounting CT root")
			logging.info("Running vzctl umount")
			proc = subprocess.Popen(
				[vzctl_bin, "--skipowner", "--skiplock", "stop", self._ctid, "--force"],
				stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
			proc_output = proc.communicate()[0]
			util.log_subprocess_output(proc_output)
			self._fs_mounted = False

	def __disable_veth(self, iface):
		logging.info("Disabling veth %s" % iface)
		proc = subprocess.Popen(
			[ip_bin, "link", "set", iface, "down"],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		# append pid's last 2 digits to serve as a "hash" which should protect us from VERY slow net namespace release
		proc = subprocess.Popen(
			[ip_bin, "link", "set", iface, "name", iface + "M" + str(os.getpid() % 100)],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)

	def migration_complete(self, fs, target_host):
		self.umount()
		fs.cleanup_shared_ploops()
		target_host.migration_complete(fs.prepare_src_data({}))
		for veth in self._veths:
			self.__disable_veth(veth.pair)

	def migration_fail(self, fs):
		fs.restore_shared_ploops()

	def target_cleanup(self, src_data):
		# Finalize restore for NFS case
		if self._has_nfs:
			self._finalize_restore()
		# Cleanup shared ploops
		if "shareds" in src_data:
			for ploop in src_data["shareds"]:
				fs_haul_ploop.merge_ploop_snapshot(
					ploop["ddxml"], ploop["guid"])

	def start(self):
		logging.info("Starting CT")
		logging.info("Running vzctl start")
		proc = subprocess.Popen(
			[vzctl_bin, "--skipowner", "--skiplock", "--ignore-ha-cluster", "start", self._ctid],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		if proc.returncode != 0:
			raise Exception("Start failed ({0})".format(proc.returncode))
		self._fs_mounted = True

	def stop(self, umount):
		logging.info("Stopping CT")
		logging.info("Running vzctl stop")
		args = [vzctl_bin, "--skipowner", "--skiplock", "stop", self._ctid]
		if not umount:
			args.append("--skip-umount")
		proc = subprocess.Popen(
			args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		self._fs_mounted = not umount

	def get_fs(self, fdfs=None, sync_copy=0):
		deltas = self.__parse_fdfs_arg(fdfs)
		return fs_haul_ploop.p_haul_fs(deltas, self._ct_priv, sync_copy)

	def get_fs_receiver(self, fdfs=None):
		deltas = self.__parse_fdfs_arg(fdfs)
		return fs_haul_ploop.p_haul_fs_receiver(deltas)

	def restored(self, pid):
		pass

	def net_lock(self):
		pass

	def net_unlock(self):
		pass

	def pre_dump(self, img):
		# Check if container has NFS mounts
		self._has_nfs = self.__has_nfs_mount()

		self.__pre_final_dump(img)

	def dump_alive(self, img):
		self.__revert_pre_dump(img)

	def run_action_scripts(self, stage):
		if stage not in vz_action_scripts:
			return
		script_env = os.environ.copy()
		script_env["VEID"] = self._ctid
		script_env["CRTOOLS_INIT_PID"] = str(self.root_task_pid())
		script_env["CRTOOLS_SCRIPT_ACTION"] = stage
		for script_bin in vz_action_scripts[stage]:
			logging.info("\t\t\tRunning action script %s" % script_bin)
			child = subprocess.Popen(script_bin, env=script_env,
				stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
			proc_output = child.communicate()[0]
			util.log_subprocess_output(proc_output)
			if child.returncode:
				raise Exception("Action script %s failed [%d]" % (script_bin, child.returncode))

	def can_migrate_tcp(self):
		return True

	def can_pre_dump(self):
		return True

	def dump_need_page_server(self):
		return True

	def __parse_fdfs_arg(self, fdfs):
		"""Parse string containing list of ploop deltas with socket fds

		String contain list of active ploop deltas with corresponding socket
		file descriptors in format %delta_path1%:%socket_fd1%[,...]. Parse it
		and return list of tuples.
		"""
		deltas = []
		for path, fd in util.parse_fdfs(fdfs):
			delta_path = fs_haul_ploop.get_delta_abspath(path, self._ct_priv)
			deltas.append((delta_path, int(fd)))

		return deltas

	def __parse_external_disks_arg(self, disks_arg):
		"""Parse multi-line string containing list of "external" disks in specific format

		Each line contain ploop disks or partition in format:
		%uuid%@%partname%:%major%:%minor%:%is_root%
		Parse it and return list of tuples.
		"""

		self._has_net_disk = False
		TUPLE_SEPARATOR = ":"
		UUID_SEPARATOR = "@"

		if not disks_arg:
			return []

		disks = []
		logging.info("Check if there are network-attached block devices")
		for disk in disks_arg.splitlines():
			try:
				uuid, rest = disk.split(UUID_SEPARATOR)
				partname, major, minor, _ = rest.split(TUPLE_SEPARATOR)
			except:
				logging.info("Unable to parse external disk argument %s" % disk)
				raise

			disks.append((uuid, int(major), int(minor)))

			if not self._has_net_disk:
				if util.is_network_disk(partname):
					logging.info(" - %s is a network-attached block device" % partname)
					self._has_net_disk = True
				else:
					logging.info(" - %s is a local block device" % partname)

		return disks


def add_hauler_args(parser):
	"""Add Virtuozzo specific command line arguments"""
	parser.add_argument("--vz-shared-disks",
						help="List of shared storage disks")
	parser.add_argument("--vz-external-disks",
						help="List of ploop disks and partitions")


def _parse_vz_config(body):
	"""Parse shell-like virtuozzo config file"""

	config_values = dict()
	for token in shlex.split(body, comments=True):
		name, sep, value = token.partition("=")
		config_values[name] = value
	return config_values


def _expand_veid_var(value, ctid):
	"""Replace shell-like VEID variable with actual container id"""
	return value.replace("$VEID", ctid).replace("${VEID}", ctid)
